import os
import hashlib
from typing import Any, Dict, Union

import boto3
import botocore


class S3:
    def __init__(
        self,
        bucket: str,
        access_policy: str = "public-read",
        prefix: Union[str, None] = None,
        encryption: bool = False,
    ):
        self.client = boto3.client("s3")
        self.bucket = bucket
        self.access_policy = access_policy
        self.prefix = prefix
        self.encryption = encryption

    def full_path(self, subpath: str):
        return os.path.join(self.prefix or "", subpath)

    def exists(self, path: str):
        try:
            return self.client.head_object(Bucket=self.bucket, Key=self.full_path(path))
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return None
            else:
                raise

    def read(self, path: str):
        try:
            response = self.client.get_object(
                Bucket=self.bucket, Key=self.full_path(path)
            )
            return response["Body"].read()
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchKey":
                return None
            else:
                raise

    def store(
        self,
        path: str,
        filename: Union[str, None] = None,
        content_type: str = "application/x-debian-package",
        cache_control: Union[str, None] = None,
        fail_if_exists: bool = False,
    ):
        if not filename:
            filename = os.path.basename(path)

        obj = self.exists(filename)

        with open(path, "rb") as f:
            content = f.read()
            file_md5 = hashlib.md5(content).hexdigest()

        # check if the object already exists
        if obj:
            if file_md5 == obj["ETag"].replace('"', "") or file_md5 == obj[
                "Metadata"
            ].get("md5", ""):
                # the same object already exists
                return True
            elif fail_if_exists:
                # explicitly raise error for duplicated filename
                raise Exception(
                    "file {} already exists with different contents".format(filename)
                )

        options: Dict[str, Any] = {
            "Bucket": self.bucket,
            "Key": self.full_path(filename),
            "ACL": self.access_policy,
            "ContentType": content_type,
            "Metadata": {"md5": file_md5},
        }
        if cache_control:
            options["CacheControl"] = cache_control

        if self.encryption:
            options["ServerSideEncryption"] = "AES256"

        with open(path, "rb") as f:
            options["Body"] = f
            try:
                self.client.put_object(**options)
                return True
            except botocore.exceptions.ClientError as e:
                if e.response["Error"]["Code"] == "404":
                    return False
                else:
                    raise



import hashlib
import os
import re
import subprocess
import tempfile
from dataclasses import dataclass, asdict, field
from os import PathLike
from typing import List, Dict, Any, Optional

import jinja2


def _escape(string: str):
    def replace(match: re.Match):
        return "%" + "".join("%{:02X}".format(c) for c in match.group(1).encode())

    return re.sub(r"([^a-zA-Z0-9_.\-~+]+)", replace, string)


@dataclass
class Package:
    name: str
    version: str
    filename: Optional[str] = field(repr=False)
    epoch: Optional[str] = field(repr=False)
    iteration: Optional[str] = field(repr=False)
    maintainer: Optional[str] = field(repr=False)
    vendor: Optional[str] = field(repr=False)
    url: Optional[str] = field(repr=False)
    category: Optional[str] = field(repr=False)
    license: Optional[str] = field(repr=False)
    architecture: Optional[str] = field(repr=False)
    description: Optional[str] = field(repr=False)
    dependencies: Optional[List[str]] = field(repr=False)
    sha1: Optional[str] = field(repr=False)
    sha256: Optional[str] = field(repr=False)
    md5: Optional[str] = field(repr=False)
    size: Optional[str] = field(repr=False)
    _url_filename: Optional[str] = field(repr=False)
    attributes: Dict[str, Any] = field(repr=False)

    @property
    def full_version(self):
        if all(v is None for v in [self.epoch, self.version, self.iteration]):
            return None

        combined_version = ":".join(filter(None, [self.epoch, self.version]))
        return (
            f"{combined_version}-{self.iteration}"
            if self.iteration
            else combined_version
        )

    def url_filename(self, codename: str):
        if self._url_filename is not None:
            return self._url_filename
        if self.filename:
            return f"pool/{codename}/{self.name[0]}/{self.name[:2]}/{os.path.basename(self.filename)}"

    def url_filename_encoded(self, codename: str):
        if self._url_filename is not None:
            return self._url_filename
        if self.filename:
            return f"pool/{codename}/{self.name[0]}/{self.name[:2]}/{_escape(os.path.basename(self.filename))}"

    @classmethod
    def parse_file(cls, filename: PathLike):
        control_data = cls.parse_control(cls.extract_control(filename))
        file_info_data = cls.parse_file_info(filename)
        package = cls(**{"filename": filename, **control_data, **file_info_data})
        return package

    @classmethod
    def parse_string(cls, string: str):
        control_data = cls.parse_control(string)
        package = cls(filename=None, **control_data)
        return package

    @staticmethod
    def extract_control(filename: PathLike) -> str:
        try:
            subprocess.run(
                ["which", "dpkg"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=True,
            )
            control_data = subprocess.check_output(["dpkg", "-f", filename]).decode(
                "utf-8"
            )
        except subprocess.CalledProcessError:
            package_files = subprocess.check_output(["ar", "t", filename]).decode(
                "utf-8"
            )
            control_filename = next(
                fname
                for fname in package_files.split("\n")
                if fname.startswith("control.")
            )
            compression = "z" if control_filename == "control.tar.gz" else "J"

            extract_control_tarball_cmd = f"ar p {filename} {control_filename}"

            try:
                subprocess.run(
                    ["ar", "t", filename, control_filename],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    check=True,
                )
            except subprocess.CalledProcessError:
                print("Failed to find control data in .deb with ar, trying tar.")
                extract_control_tarball_cmd = (
                    f"tar {compression}xf {filename} --to-stdout {control_filename}"
                )

            with tempfile.TemporaryDirectory() as path:
                extract_cmd = (
                    f"{extract_control_tarball_cmd} | tar -{compression}xf - -C {path}"
                )
                subprocess.run(extract_cmd, shell=True, check=True)
                with open(
                    os.path.join(path, "control"), "r", encoding="utf-8"
                ) as control_file:
                    control_data = control_file.read()

        return control_data

    @classmethod
    def parse_control(cls, control) -> Dict[str, Any]:
        # Gather fields in control file
        fields = {}
        for line in control.splitlines():
            field = None
            value = ""
            if match := re.match(r"^(\s+)(\S.*)$", line):
                indent = match.group(1)
                rest = match.group(2)

                if len(indent) == 1 and rest == ".":
                    value += "\n"
                    rest = ""
                elif len(value) > 0:
                    value += "\n"

                value += rest
            elif match := re.match(r"^([-\w]+):(.*)$", line):
                if field:
                    fields[field] = value

                field = match.group(1)
                value = match.group(2).strip()

            if field:
                fields[field] = value

        # Packages manifest fields
        data = {}

        # Parse 'epoch:version-iteration' in the version string
        full_version = fields.pop("Version")
        version_match = re.match(r"^(?:([0-9]+):)?(.+?)(?:-(.*))?$", full_version)
        if not version_match:
            raise ValueError(f"Unsupported version string '{full_version}'")

        epoch, version, iteration = version_match.groups()
        data["epoch"] = epoch
        data["version"] = version
        data["iteration"] = iteration

        data["architecture"] = fields.pop("Architecture", None)
        data["category"] = fields.pop("Section", None)
        data["license"] = fields.pop("License", None) or "unknown"
        data["maintainer"] = fields.pop("Maintainer", None)
        data["name"] = fields.pop("Package", None)
        data["url"] = fields.pop("Homepage", None)
        data["vendor"] = fields.pop("Vendor", None) or "none"
        data["_url_filename"] = fields.pop("Filename", None)
        data["sha1"] = fields.pop("SHA1", None)
        data["sha256"] = fields.pop("SHA256", None)
        data["md5"] = fields.pop("MD5sum", None)
        data["size"] = fields.pop("Size", None)
        data["description"] = fields.pop("Description", None)
        data["dependencies"] = cls.parse_depends(fields.pop("Depends", None))

        # other attributes
        attributes: Dict[str, Any] = {}
        attributes["deb_priority"] = fields.pop("Priority", None)
        attributes["deb_origin"] = fields.pop("Origin", None)
        attributes["deb_installed_size"] = fields.pop("Installed-Size", None)
        attributes["deb_recommends"] = fields.pop("Recommends", None)
        attributes["deb_suggests"] = fields.pop("Suggests", None)
        attributes["deb_enhances"] = fields.pop("Enhances", None)
        attributes["deb_pre_depends"] = fields.pop("Pre-Depends", None)
        attributes["deb_breaks"] = fields.pop("Breaks", None)
        attributes["deb_conflicts"] = fields.pop("Conflicts", None)
        attributes["deb_provides"] = fields.pop("Provides", None)
        attributes["deb_replaces"] = fields.pop("Replaces", None)
        attributes["deb_field"] = {k.split("-", 1)[-1]: v for k, v in fields.items()}
        data["attributes"] = attributes

        return data

    @staticmethod
    def parse_depends(data):
        if data is None or not data:
            return []

        # parse dependencies. Debian dependencies come in one of two forms:
        # * name
        # * name (op version)
        # They are all on one line, separated by ", "

        dep_re = re.compile(r"^([^ ]+)(?: \(([>=<]+) ([^)]+)\))?$")
        dependencies = []

        for dep in re.split(r", *", data):
            m = dep_re.match(dep)
            if m:
                name, op, version = m.groups()
                # this is the proper form of dependency
                if op and version and op != "" and version != "":
                    dependencies.append(f"{name} ({op} {version})".strip())
                else:
                    dependencies.append(name.strip())
            else:
                # Assume normal form dependency, "name op version".
                dependencies.append(dep)

        return dependencies

    @staticmethod
    def parse_file_info(filename: PathLike):
        data: Dict[str, Any] = {}
        data["size"] = os.path.getsize(filename)

        with open(filename, "rb") as f:
            content = f.read()
            data["sha1"] = hashlib.sha1(content).hexdigest()
            data["sha256"] = hashlib.sha256(content).hexdigest()
            data["md5"] = hashlib.md5(content).hexdigest()

        return data

    def generate(self, codename: str):
        env = jinja2.Environment(
            loader=jinja2.PackageLoader("dpkgs3"),
            autoescape=jinja2.select_autoescape(),
        )
        template = env.get_template("package.jinja2")
        return template.render(
            **asdict(self),
            url_filename_encoded=self.url_filename_encoded(codename),
        )



import gzip
import hashlib
import os
import tempfile
from dataclasses import dataclass, field
from typing import Any, List, Dict, Optional

from dpkgs3.package import Package
from dpkgs3.s3 import S3


@dataclass
class Manifest:
    codename: str
    component: str
    architecture: str
    packages: List[Package] = field(default_factory=list, repr=False)
    packages_to_be_uploaded: List[Package] = field(default_factory=list, repr=False)
    files: Dict[str, str] = field(default_factory=dict, repr=False)
    cache_control: str = field(default="", repr=False)
    fail_if_exists: bool = field(default=False, repr=False)
    skip_package_upload: bool = field(default=False, repr=False)

    @classmethod
    def retrieve(
        cls, s3: S3, codename: str, component: str, architecture: str, **kwargs
    ):
        m = cls(
            codename,
            component,
            architecture,
            **kwargs,
        )
        packages_data = s3.read(
            f"dists/{codename}/{component}/binary-{architecture}/Packages"
        )
        if packages_data is not None:
            m.packages = cls.parse_packages(packages_data.decode("utf-8"))

        return m

    @staticmethod
    def parse_packages(packages_data: str) -> List[Package]:
        packages = []
        for s in packages_data.split("\n\n"):
            if s.rstrip():
                packages.append(Package.parse_string(s))
        return packages

    def add_package(
        self,
        pkg: Package,
        preserve_versions: bool = True,
        fail_if_exists: bool = True,
        needs_uploading: bool = True,
    ):
        if fail_if_exists:
            for p in self.packages:
                if (
                    p.name == pkg.name
                    and p.full_version == pkg.full_version
                    and os.path.basename(p.url_filename(self.codename))
                    == os.path.basename(pkg.url_filename(self.codename))
                ):
                    raise Exception(
                        f"package {pkg.name}_{pkg.full_version} already exists with filename ({p.url_filename(self.codename)})"
                    )

        self.packages = [
            p
            for p in self.packages
            if not (
                p.name == pkg.name
                and (p.full_version == pkg.full_version if preserve_versions else True)
            )
        ]
        self.packages.append(pkg)
        if needs_uploading:
            self.packages_to_be_uploaded.append(pkg)

        return pkg

    def delete_package(
        self,
        pkg_name: str,
        versions: Optional[List[str]] = None,
    ):
        deleted_packages = []
        versions = versions if versions is not None else []
        for p in self.packages:
            if p.name == pkg_name and any(
                [
                    possible_version in versions
                    for possible_version in [
                        p.version,
                        p.full_version,
                        f"{p.version}-{p.iteration}",
                    ]
                ]
            ):
                deleted_packages.append(p)

        self.packages = list(set(self.packages) - set(deleted_packages))
        return deleted_packages

    def generate(self):
        return "\n".join([package.generate(self.codename) for package in self.packages])

    def write_to_s3(self, s3: S3):
        manifest = self.generate()

        if not self.skip_package_upload:
            for pkg in self.packages_to_be_uploaded:
                assert pkg.filename is not None
                s3.store(
                    pkg.filename,
                    pkg.url_filename(self.codename),
                    "application/x-debian-package",
                    self.cache_control,
                    fail_if_exists=self.fail_if_exists,
                )

        # generate the Packages file
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as pkgs_temp:
            pkgs_temp.write(manifest)
            pkgs_temp.close()
            f = f"dists/{self.codename}/{self.component}/binary-{self.architecture}/Packages"
            s3.store(pkgs_temp.name, f, "text/plain; charset=utf-8", self.cache_control)
            self.files[
                f"{self.component}/binary-{self.architecture}/Packages"
            ] = self.hashfile(pkgs_temp.name)
            os.unlink(pkgs_temp.name)

        # generate the Packages.gz file
        with tempfile.NamedTemporaryFile(delete=False) as gztemp:
            gztemp.close()
            with gzip.open(gztemp.name, "wb") as gz:
                gz.write(manifest.encode("utf-8"))
            f = f"dists/{self.codename}/{self.component}/binary-{self.architecture}/Packages.gz"
            s3.store(gztemp.name, f, "application/x-gzip", self.cache_control)
            self.files[
                f"{self.component}/binary-{self.architecture}/Packages.gz"
            ] = self.hashfile(gztemp.name)
            os.unlink(gztemp.name)

    @staticmethod
    def hashfile(path: str):
        data: Dict[str, Any] = {}
        data["size"] = os.path.getsize(path)

        with open(path, "rb") as f:
            content = f.read()
            data["sha1"] = hashlib.sha1(content).hexdigest()
            data["sha256"] = hashlib.sha256(content).hexdigest()
            data["md5"] = hashlib.md5(content).hexdigest()

        return data



import os
from pathlib import Path
from typing import Dict, List, Optional

import click
from dpkgs3.s3 import S3
from dpkgs3.release import Release
from dpkgs3.manifest import Manifest
from dpkgs3.package import Package


def configure_s3_client(
    s3_bucket: str,
    s3_prefix: Optional[str],
    s3_visibility: str,
    s3_encryption: bool,
    aws_default_region: Optional[str],
    aws_access_key_id: Optional[str],
    aws_secret_access_key: Optional[str],
):
    if aws_default_region is not None:
        os.environ.setdefault("AWS_DEFAULT_REGION", aws_default_region)
    if aws_access_key_id is not None:
        os.environ.setdefault("AWS_ACCESS_KEY_ID", aws_access_key_id)
    if aws_secret_access_key is not None:
        os.environ.setdefault("AWS_SECRET_ACCESS_KEY", aws_secret_access_key)

    s3 = S3(
        bucket=s3_bucket,
        prefix=s3_prefix,
        access_policy=s3_visibility,
        encryption=s3_encryption,
    )
    return s3


def add_common_options():
    common_options = [
        click.option(
            "--codename",
            default="stable",
            show_default=True,
            help="The codename of the APT repository",
        ),
        click.option(
            "--component",
            default="main",
            show_default=True,
            help="The component of the APT repository",
        ),
        click.option("--s3-bucket", required=True, help="S3 bucket name"),
        click.option(
            "--s3-prefix",
            default="",
            show_default=True,
            help="Path prefix to use when storing on S3",
        ),
        click.option(
            "--s3-visibility",
            type=click.Choice(
                choices=[
                    "public-read",
                    "private",
                    "authenticated-read",
                    "bucket-owner-full-control",
                ]
            ),
            default="public-read",
            show_default=True,
            help="The access policy for the uploaded files",
        ),
        click.option(
            "--s3-encryption",
            is_flag=True,
            default=False,
            show_default=True,
            help="Use S3 server side encryption",
        ),
        click.option("--aws-default-region", default=None, help="AWS default region"),
        click.option("--aws-access-key-id", default=None, help="AWS Access Key ID"),
        click.option(
            "--aws-secret-access-key", default=None, help="AWS Secret Access Key"
        ),
    ]

    def inner(func):
        for option in reversed(common_options):
            func = option(func)
        return func

    return inner


@click.group()
def cli(**kwargs):
    pass


@cli.command()
@add_common_options()
@click.option(
    "--preserve-versions",
    is_flag=True,
    default=True,
    show_default=True,
    help="Whether to preserve other versions of a package in the repository when uploading one",
)
@click.option(
    "--cache-control", default="", help="Cache-Control headers to add to S3 objects"
)
@click.option(
    "--fail-if-exists",
    is_flag=True,
    default=False,
    show_default=True,
    help="Whether to preserve other versions of a package in the repository when uploading one",
)
@click.option(
    "--skip-package-upload",
    is_flag=True,
    default=False,
    show_default=True,
    help="Whether to skip all package uploads. This is useful when hosting .deb outside of the bucket.",
)
@click.argument(
    "files", nargs=-1, type=click.Path(exists=True, file_okay=True, dir_okay=False)
)
def upload(
    files: List[Path],
    codename: str,
    component: str,
    preserve_versions: bool,
    cache_control: str,
    fail_if_exists: bool,
    skip_package_upload: bool,
    s3_bucket: str,
    s3_prefix: str,
    s3_visibility: str,
    s3_encryption: bool,
    aws_default_region: Optional[str],
    aws_access_key_id: Optional[str],
    aws_secret_access_key: Optional[str],
):
    if len(files) == 0:
        raise click.MissingParameter(param_type="argument", param_hint="FILES")

    for filename in files:
        _, ext = os.path.splitext(filename)
        if ext != ".deb":
            raise click.BadParameter(
                param_hint="FILES", message=f"{filename} is not a .deb file"
            )

    s3 = configure_s3_client(
        s3_bucket,
        s3_prefix,
        s3_visibility,
        s3_encryption,
        aws_default_region,
        aws_access_key_id,
        aws_secret_access_key,
    )

    print("Retrieving existing manifests")
    release = Release.retrieve(
        s3=s3,
        codename=codename,
        cache_control=cache_control,
    )
    manifests: Dict[str, Manifest] = {}
    for arch in release.architectures:
        manifests[arch] = Manifest.retrieve(
            s3=s3,
            codename=codename,
            component=component,
            architecture=arch,
            cache_control=cache_control,
            skip_package_upload=skip_package_upload,
        )

    for filename in files:
        pkg = Package.parse_file(filename)
        manifest = manifests[pkg.architecture]
        print(f"Add package file {filename} to {manifest}")
        try:
            manifest.add_package(
                pkg, preserve_versions=preserve_versions, fail_if_exists=fail_if_exists
            )
        except Exception as e:
            raise click.ClickException(str(e))

    for manifest in manifests.values():
        if len(manifest.packages_to_be_uploaded) > 0:
            print(f"Uploading packages in {manifest}")
            manifest.write_to_s3(s3=s3)
            release.update_manifest(manifest)

    print("Updating Release file")
    release.write_to_s3(s3=s3)

    print("Update complete.")



import datetime
import os
import re
import tempfile
from dataclasses import dataclass, asdict, field
from typing import Any, List, Union, Dict, Optional

import jinja2
from dpkgs3.s3 import S3
from dpkgs3.manifest import Manifest


@dataclass
class Release:
    codename: str
    origin: Optional[str] = field(default=None, repr=False)
    suite: Optional[str] = field(default=None, repr=False)
    architectures: List[str] = field(default_factory=list, repr=False)
    components: List[str] = field(default_factory=list, repr=False)
    cache_control: Optional[str] = field(default=None, repr=False)
    files: Dict[str, Any] = field(default_factory=dict, repr=False)
    access_policy: str = field(default="public-read", repr=False)

    @classmethod
    def retrieve(
        cls,
        s3: S3,
        codename: str,
        origin: Optional[str] = None,
        suite: Union[str, None] = None,
        cache_control: Union[str, None] = None,
    ):
        release_data = s3.read(f"dists/{codename}/Release")
        if release_data is not None:
            rel = cls.parse_release(release_data.decode("utf-8"))
        else:
            rel = cls(codename=codename)

        rel.codename = codename
        rel.origin = origin if origin is not None else rel.origin
        rel.suite = suite if suite is not None else rel.suite
        rel.cache_control = cache_control

        return rel

    @classmethod
    def parse_release(cls, release_data: str):
        lines = release_data.split("\n")

        def parse_field(field):
            value = next(
                (
                    line.split(": ", 1)[1]
                    for line in lines
                    if line.startswith(field + ": ")
                ),
                None,
            )
            return value

        # Initialize variables to None or empty lists/dicts
        codename = parse_field("Codename")
        origin = parse_field("Origin")
        suite = parse_field("Suite")
        architectures = re.findall(r"^Architectures: (.+)", release_data, re.MULTILINE)
        architectures = architectures[0].split() if architectures else []
        components = re.findall(r"^Components: (.+)", release_data, re.MULTILINE)
        components = components[0].split() if components else []

        files: Dict[str, Any] = {}

        # Find all the hashes
        hashes = re.findall(r"^\s+([^\s]+)\s+(\d+)\s+(.+)$", release_data, re.MULTILINE)
        for hash, size, name in hashes:
            if name in files:
                files[name].update({"size": int(size)})
            else:
                files[name] = {"size": int(size)}

            if len(hash) == 32:
                files[name]["md5"] = hash
            elif len(hash) == 40:
                files[name]["sha1"] = hash
            elif len(hash) == 64:
                files[name]["sha256"] = hash

        return cls(
            codename=codename,
            origin=origin,
            suite=suite,
            architectures=architectures,
            components=components,
            files=files,
        )

    def update_manifest(self, manifest: Manifest):
        if manifest.component not in self.components:
            self.components.append(manifest.component)
        if manifest.architecture not in self.architectures:
            self.architectures.append(manifest.architecture)
        self.files.update(manifest.files)

    def validate_others(self):
        to_apply = []
        for comp in self.components:
            for arch in ["amd64", "i386", "armhf"]:
                if f"{comp}/binary-{arch}/Packages" not in self.files:
                    m = Manifest(
                        codename=self.codename, component=comp, architecture=arch
                    )
                    m.write_to_s3()
                    to_apply.append(m)

        for m in to_apply:
            self.update_manifest(m)

    def generate(self):
        env = jinja2.Environment(
            loader=jinja2.PackageLoader("dpkgs3"),
            autoescape=jinja2.select_autoescape(),
        )
        template = env.get_template("release.jinja2")
        return template.render(
            **asdict(self),
            now=datetime.datetime.now()
            .astimezone(datetime.timezone.utc)
            .strftime("%a, %d %b %Y %T %Z"),
        )

    def write_to_s3(self, s3: S3):
        self.validate_others()

        with tempfile.NamedTemporaryFile(mode="w", delete=False) as release_tmp:
            release_tmp.write(self.generate())
            release_tmp.close()

            s3.store(
                release_tmp.name,
                f"dists/{self.codename}/Release",
                "text/plain; charset=utf-8",
                self.cache_control,
            )

            # TODO: gpg signing key

            os.unlink(release_tmp.name)
